import os
import time
from datetime import datetime
from pathlib import Path
from typing import Dict
from uuid import uuid4

from adb_pywrapper.adb_device import AdbDevice
from adb_pywrapper.adb_screen_recorder import AdbScreenRecorder
from appium.options.android import UiAutomator2Options
from appium.webdriver import WebElement
from appium.webdriver.common.appiumby import AppiumBy
from appium import webdriver
from appium.webdriver.extensions.android.nativekey import AndroidKey
from appium.webdriver.webdriver import WebDriver
from selenium.webdriver import ActionChains
from urllib3.exceptions import MaxRetryError

from puma.computer_vision import ocr
from puma.computer_vision.ocr import RecognizedText
from puma.state_graph import logger
from puma.utils import CACHE_FOLDER
from puma.utils.gtl_logging import create_gtl_logger

# Keycode constants, found at
KEYCODE_LEFT_ARROW = 21
KEYCODE_ENTER = 66
KEYCODE_BACKSPACE = 67

class PumaClickException(Exception):
    """
    Custom exception for handling errors related to clicking actions in the PumaDriver.
    """
    pass

def _get_android_default_options() -> UiAutomator2Options:
    """
    Creates and configures default options for an Android UiAutomator2 driver.

    This function sets up the default options required for initializing an Android
    UiAutomator2 driver, including platform name and command timeout settings.

    :return: Configured UiAutomator2Options instance.
    """
    options = UiAutomator2Options()
    options.no_reset = True
    options.platform_name = 'Android'
    options.new_command_timeout = 1200
    return options


__drivers: dict[str, WebDriver] = {}

def _get_appium_driver(appium_server: str, udid: str, options) -> WebDriver:
    key = f"{appium_server}${udid}"
    if key not in __drivers.keys():
        try:
            __drivers[key] = webdriver.Remote(appium_server, options=options)
        except MaxRetryError:
            logger.error("Connecting to the Appium server has failed.\n"
                         "Make sure that the appium server is running!\n"
                         "This can be done by running the `appium` command from the command line.")
            exit(1)
    else:
        logger.warning(f'WARNING: There already was an initialized driver for appium server {appium_server} and udid {udid}. '
                       'This driver will be used, which might mean your Appium capabilities are ignored as these cannot be'
                       'altered for a driver that has already been initialized. If you need specific capabilities, please '
                       'rewrite your Puma code to ensure the correct capabilities are loaded the first time you connect to '
                       f'server {appium_server} and device {udid}.')
    return __drivers[key]

def supported_version(version: str):
    def decorator(class_or_function):
        class_or_function.supported_version = version
        return class_or_function

    return decorator

class PumaDriver:
    """
    A driver class for interacting with Android applications using Appium.

    This class provides methods to interact with an Android app, such as activating,
    terminating, and interacting with UI elements. It uses Appium's WebDriver for
    remote control of the application.
    """

    def __init__(self, udid: str, app_package: str, implicit_wait: int = 1, appium_server: str = 'http://localhost:4723', desired_capabilities: Dict[str, str] = None):
        """
        Initializes the PumaDriver with device and application details.

        :param udid: The unique device identifier for the Android device.
        :param app_package: The package name of the application to interact with.
        :param implicit_wait: The implicit wait time for element searches, defaults to 1 second.
        :param appium_server: The address of the Appium server, defaults to 'http://localhost:4723'.
        :param desired_capabilities: The desired capabilities as passed to the Appium webdriver.
        """
        self.options = _get_android_default_options()
        self.options.udid = udid
        self.app_package = app_package
        if desired_capabilities:
            self.options.load_capabilities(desired_capabilities)
        logger.info("Connecting to Appium driver...")
        self.driver = _get_appium_driver(appium_server, udid, self.options)
        self.implicit_wait = implicit_wait
        self.driver.implicitly_wait(implicit_wait)
        self.udid = self.driver.capabilities.get("udid")
        self.adb = AdbDevice(self.udid)
        self._screen_recorder = None
        self._screen_recorder_output_directory = None
        self.gtl_logger = create_gtl_logger(udid)

    def is_present(self, xpath: str, implicit_wait: float = 0) -> bool:
        """
        Checks if an element is present on the screen.

        :param xpath: The XPath of the element to check.
        :param implicit_wait: The time to wait for the element to be present.
        :return: True if the element is present, False otherwise.
        """
        self.driver.implicitly_wait(implicit_wait)
        found = self.driver.find_elements(by=AppiumBy.XPATH, value=xpath)
        self.driver.implicitly_wait(self.implicit_wait)
        return len(found) > 0

    def activate_app(self):
        """
        Activates the application on the device.
        """
        self.gtl_logger.info(f'Activating app {self.app_package}')
        self.driver.activate_app(self.app_package)

    def terminate_app(self):
        """
        Terminates the application on the device.
        """
        self.gtl_logger.info(f'Terminating app {self.app_package}')
        self.driver.terminate_app(self.app_package)

    def restart_app(self):
        """
        Restarts the application by terminating and then activating it.
        """
        self.terminate_app()
        self.activate_app()

    def app_open(self) -> bool:
        """
        Checks if the application is currently open.

        :return: True if the application is open, False otherwise.
        """
        return str(self.driver.current_package) == self.app_package

    def back(self):
        """
        Simulates pressing the back button on the device.
        """
        self.gtl_logger.info(f'Pressing back button')
        self.driver.press_keycode(AndroidKey.BACK)

    def home(self):
        """
        Simulates pressing the home button on the device.
        """
        self.gtl_logger.info(f'Pressing home button')
        self.driver.press_keycode(AndroidKey.HOME)

    def click(self, xpath: str, width_ratio:float=0.5, height_ratio:float=0.5):
        """
        Clicks on an element specified by its XPath.

        By default, this method clicks in the center of the element selected by the given XPath.
        If you want to click off-center, you can use the width and heigh ratio.
        The ratios are values between 0 and 1 that determine where the element needs to be clicked, where (0,0)
        corresponds to the top-left and (1,1) corresponds to the bottom right.
        The width_ratio determines the x coordinate, the height_ratio the y coordinate.

        :param xpath: The XPath of the element to click.
        :param width_ratio: Optional. Determines the x coordinate, relative within the element, from 0 to 1 (left to right).
        :param height_ratio: Optional. Determines the y coordinate, relative within the element, from 0 to 1 (top to bottom).
        :raises PumaClickException: If the element cannot be clicked after multiple attempts.
        """
        for attempt in range(3):
            if self.is_present(xpath, self.implicit_wait):
                if (width_ratio, height_ratio) == (0.5, 0.5):
                    self.driver.find_element(by=AppiumBy.XPATH, value=xpath).click()
                else:
                    element = self.get_element(xpath)
                    top_left = element.location['x'], element.location['y']
                    size = element.size['height'], element.size['width']
                    location = int(top_left[0] + width_ratio * size[1]), int(top_left[1] + height_ratio * size[0])
                    self.tap(location)
                return
        raise PumaClickException(f'Could not click on non present element with xpath {xpath}')

    def tap(self, coords: tuple[int, int]):
        """
        Taps on the screen at the specified coordinates.

        :param coords: A tuple (x, y) representing the coordinates to tap.
        """
        self.gtl_logger.info(f'Tapping on coordinates {coords}')
        self.driver.tap([coords])

    def long_click_element(self, xpath: str, duration: int = 1):
        """
        Clicks on a certain element, and hold for a given duration (in seconds)

        :param xpath: The XPath of the element to click.
        :param duration: how many seconds to hold the element before releasing
        :raises PumaClickException: If the element cannot be found after multiple attempts.
        """
        element = self.get_element(xpath)
        actions = ActionChains(self.driver)
        actions.move_to_element(element).click_and_hold().pause(duration).release().perform()

    def get_element(self, xpath: str):
        """
        Retrieves an element specified by its XPath.

        :param xpath: The XPath of the element to retrieve.
        :return: The WebElement corresponding to the XPath.
        :raises PumaClickException: If the element cannot be found after multiple attempts.
        """
        for attempt in range(3):
            if self.is_present(xpath, self.implicit_wait):
                return self.driver.find_element(by=AppiumBy.XPATH, value=xpath)
        raise PumaClickException(f'Could not find element with xpath {xpath}')

    def get_elements(self, xpath: str) -> list[WebElement]:
        """
        Retrieves all elements matching the specified XPath.

        :param xpath: The XPath of the elements to retrieve.
        :return: A list of WebElements corresponding to the XPath.
        :raises PumaClickException: If no elements can be found after multiple attempts.
        """
        for attempt in range(3):
            if self.is_present(xpath, self.implicit_wait):
                return self.driver.find_elements(by=AppiumBy.XPATH, value=xpath)
        raise PumaClickException(f'Could not find elements with xpath {xpath}')

    def _swipe_down(self):
        window_size = self.driver.get_window_size()
        start_x = window_size['width'] / 2
        start_y = window_size['height'] * 0.8
        end_y = window_size['height'] * 0.2
        self.driver.swipe(start_x, start_y, start_x, end_y, 500)
        time.sleep(0.5)

    def swipe_to_find_element(self, xpath: str, max_swipes: int = 10):
        """
        Swipes down to find an element specified by its XPath. This is necessary when the element you want to click on
        is out of view.

        :param xpath: The XPath of the element to find.
        :param max_swipes: The maximum number of swipe attempts to find the element.
        :raises PumaClickException: If the element cannot be found after the maximum number of swipes.
        """
        for attempt in range(max_swipes):
            if self.is_present(xpath):
                return self.get_element(xpath)
            else:
                self.gtl_logger.warning(f"Attempt {attempt + 1}: Element not found, swiping down")
                self._swipe_down()
        raise PumaClickException(f'After {max_swipes} swipes, cannot find element with xpath {xpath}')

    def swipe_to_find_elements(self, xpath: str, num_swipes: int = 10):
        """
        Collects all elements matching given XPath. Next, swipes down a number of times and collects
        all new elements matching given XPath in each resulting view.

        :param xpath: the xpath of the elements to find
        :param num_swipes: the number of swipes to execute
        :raises PumaClickException: if no matching element can be found after given number of swipes
        """
        seen_elements = set()
        results = []

        for attempt in range(num_swipes):
            try:
                found_elements = self.get_elements(xpath)
            except PumaClickException as e:
                self.gtl_logger.info(f'Unable to get elements with XPath {xpath}, reason: {str(e)}')
                self.gtl_logger.info(f'Swiping down')
            else:
                for element in found_elements:
                    if element not in seen_elements:
                        seen_elements.add(element)
                        results.append(element)

            self._swipe_down()

        if not results:
            raise PumaClickException(f'After {num_swipes} swipes, no element with xpath {xpath} found')

        return results

    def swipe_to_click_element(self, xpath: str, max_swipes: int = 10):
        """
        Swipes down to find and click an element specified by its XPath. This is necessary when the element you want to
        click on is out of view.

        :param xpath: The XPath of the element to find and click.
        :param max_swipes: The maximum number of swipe attempts to find the element.
        :raises PumaClickException: If the element cannot be found after the maximum number of swipes.
        """
        self.swipe_to_find_element(xpath, max_swipes)
        self.click(xpath)

    def send_keys(self, xpath: str, text: str):
        """
        Sends keys to an element specified by its XPath.

        :param xpath: The XPath of the element to send keys to.
        :param text: The text to send to the element.
        """
        self.gtl_logger.info(f'Entering text "{text}" in text box')
        element = self.get_element(xpath)
        element.click() # TODO check all usages
        time.sleep(0.5)
        # The element has changed after clicking due to the keyboard appearing, so find it again.
        element = self.get_element(xpath)
        element.clear()
        element.send_keys(text)

    def press_enter(self):
        """
        Presses the ENTER key.
        """
        self.driver.press_keycode(KEYCODE_ENTER)

    def press_backspace(self):
        """
        Presses the BACKSPACE key.
        """
        self.driver.press_keycode(KEYCODE_BACKSPACE)

    def press_left_arrow(self):
        """
        Presses the LEFT ARROW key.
        """
        self.driver.press_keycode(KEYCODE_LEFT_ARROW)

    def open_url(self, url: str):
        """
        Opens a given URL. The URl will open in the default app configured for that URL.
        """
        self.adb.open_intent(url)

    def open_notifications(self):
        """
        Opens the Android notifications panel.
        """
        self.gtl_logger.info('Opening notifications panel')
        self.driver.open_notifications()

    def start_recording(self, output_directory: str):
        """
        Starts a screen recording.

        :param output_directory: The directory the screen recording should be stored in.
        """
        if self._screen_recorder is None:
            self._screen_recorder_output_directory = output_directory
            self._screen_recorder = AdbScreenRecorder(self.adb)
            self.gtl_logger.info('Starting screen recording')
            self._screen_recorder.start_recording()

    def stop_recording_and_save_video(self) -> list[str] | None:
        if self._screen_recorder is None:
            return None
        self.gtl_logger.info('Ending screen recording')
        video_files = self._screen_recorder.stop_recording(self._screen_recorder_output_directory)
        self._screen_recorder.__exit__(None, None, None)
        self._screen_recorder = None
        return video_files

    def _new_screenshot_name(self):
        now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        device_name = self.options.device_name
        return Path(CACHE_FOLDER) / f'{now}-{device_name}-{uuid4()}.png'

    def _find_text_ocr(self, text_to_find: str) -> list[RecognizedText]:
        path = self._new_screenshot_name()
        screenshot_taken = False
        try:
            screenshot_taken = self.driver.get_screenshot_as_file(path)
            if not screenshot_taken:
                raise Exception(f'Screenshot could not be stored to {path}')
            self.gtl_logger.info(f'Using OCR to find text "{text_to_find}"')
            found_text = ocr.find_text(str(path), text_to_find)
            return found_text
        finally:
            if screenshot_taken:
                os.remove(path)

    def click_text_ocr(self, text_to_click: str, click_first_when_multiple: bool = False):
        """
        Clicks a text if it can be found on a screen using OCR.

        :param text_to_click: The text to click.
        :param click_first_when_multiple: If True, the first occurrence of the string will be clicked if multiple are found.
        If False, raises an PumaClickException if multiple occurrences are found. Defaults to False.
        """
        self.gtl_logger.info(f'Using OCR to click on text "{text_to_click}"')
        found_text = self._find_text_ocr(text_to_click)
        if len(found_text) == 0:
            msg = f'Could not find text "{text_to_click}" on screen so could not click it'
            raise PumaClickException(msg)
        if len(found_text) > 1:
            msg = f'Found multiple occurrences of text "{text_to_click}" on screen so could not determine what to click'
            if not click_first_when_multiple:
                raise PumaClickException(msg)
            else:
                self.gtl_logger.warning(f'Found multiple occurrences of text "{text_to_click}" on screen, clicking first one')
        x = found_text[0].bounding_box.middle[0]
        y = found_text[0].bounding_box.middle[1]
        self.gtl_logger.info(f'Clicking found text "{found_text}" at coordinates {(x,y)}')
        self.driver.execute_script('mobile: clickGesture', {'x': x, 'y': y})

    def set_idle_timeout(self, timeout: int):
        """
        Sets a maximum time to wait while idle.

        :param timeout: The maximum time to wait.
        """
        # https://github.com/appium/appium-uiautomator2-driver#poor-elements-interaction-performance
        # https://github.com/appium/appium-uiautomator2-driver#settings-api
        settings = self.driver.get_settings()
        settings.update({"waitForIdleTimeout": timeout})
        self.driver.update_settings(settings)

    def execute_script(self, script: str):
        self.driver.execute_script(script)

    def __repr__(self):
        """
        Returns a string representation of the PumaDriver instance.

        :return: A string describing the PumaDriver instance.
        """
        return f"Puma Driver {self.options.udid} for app package {self.app_package}"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._screen_recorder is not None:
            self.stop_recording_and_save_video()
        self.driver.__exit__(exc_type, exc_val, exc_tb)


from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass
from typing import Callable, List

from puma.state_graph import logger
from puma.state_graph.puma_driver import PumaDriver


class State(ABC):
    """
    Abstract class representing a state. Each state represents a window in the UI.
    """

    def __init__(self, initial_state: bool = False, parent_state: 'State' = None, parent_state_transition: Callable[..., None] = None):
        """
        Initializes a new State instance.

        :param initial_state: Whether this is the initial state of the FSM.
        :param parent_state: The parent state of this state, or None if it has no parent.
        :param parent_state_transition: How to transition back to the parent state. By default, this is a press on the back button.
        """
        self.id = None  # set in metaclass
        self.initial_state = initial_state
        self.parent_state = parent_state
        self.transitions = []

        if parent_state:
            if parent_state_transition is None:
                self.to(parent_state, back)
            else:
                self.to(parent_state, parent_state_transition)

    def to(self, to_state: 'State', ui_actions: Callable[..., None]):
        """
        Define the transition from this state to another state.

        :param to_state: The next state to transition to.
        :param ui_actions: A list of UI action functions to perform the transition.
        """
        self.transitions.append(Transition(self, to_state, ui_actions))

    def from_states(self, from_states: List['State'], ui_actions: Callable[..., None]):
        """
        Define the transition from a set of other states to this state.
        This method is convenient when a state can be reached from many other states with the same UI actions.

        :param from_states: The next state to transition to.
        :param ui_actions: A list of UI action functions to perform the transition.
        """
        for state in from_states:
            state.to(self, ui_actions)

    @abstractmethod
    def validate(self, driver: PumaDriver) -> bool:
        """
        Abstract method to validate the state.

        :param driver: The PumaDriver instance to use.
        """
        pass

    def __repr__(self):
        return f"{self.id}"


class ContextualState(State):
    @abstractmethod
    def validate_context(self, driver: PumaDriver) -> bool:
        """
        Abstract method to validate the contextual state.

        :param driver: The PumaDriver instance to use.
        """
        pass


class SimpleState(State):
    """
    Simple State. This is a standard state which can be validated by providing a list of present XPaths.
    """

    def __init__(self, xpaths: List[str], invalid_xpaths: list[str] = [], initial_state: bool = False, parent_state: 'State' = None, parent_state_transition: Callable[..., None] = None):
        """
        Initializes a new SimpleState instance.

        :param xpaths: A list of XPaths which are all present on the state window.
        :param invalid_xpaths: A list of xpaths which cannot be present in the state window.
        :param initial_state: Whether this is the initial state.
        :param parent_state: The parent state of this state, or None if it has no parent.
        :param parent_state_transition: How to transition back to the parent state. By default, this is a press on the back button.
        """
        super().__init__(initial_state=initial_state, parent_state=parent_state, parent_state_transition=parent_state_transition)
        if not xpaths:
            raise ValueError(f'Cannot create a SimpleState without any xpath validation expressions.')
        self.present_xpaths = xpaths
        self.invalid_xpaths = invalid_xpaths

    def validate(self, driver: PumaDriver) -> bool:
        """
        Validates if all XPaths are present on the screen.

        :param driver: The PumaDriver instance to use.
        :return: True if all XPaths are present, otherwise False.
        """
        return (all(driver.is_present(xpath) for xpath in self.present_xpaths)
                and all((not driver.is_present(xpath)) for xpath in self.invalid_xpaths))


def back(driver: PumaDriver):
    """
    Utility method for calling the back action in Android devices.

    :param driver: The PumaDriver instance to use.
    """
    logger.info(f'calling driver.back() with driver {driver}')
    driver.back()

@dataclass
class Transition:
    """
    A class representing a transition between states.

    This class encapsulates the details of a transition, including the starting state,
    the destination state, and any associated UI actions that should be executed
    to perform the transition.

    :param from_state: The starting state of the transition.
    :param to_state: The destination state of the transition.
    :param ui_actions: A function to be called with optional arguments during the transition,
                        typically to perform UI-related actions.
    """
    from_state: State
    to_state: State
    ui_actions: Callable[..., None]


def compose_clicks(xpaths: List[str], name: str = 'click') -> Callable[[PumaDriver], None]:
    """
    Helper function to create a lambda for constructing transitions by clicking elements.

    This function generates a lambda function that, when executed, will click on a series
    of elements specified by their XPaths.

    :param xpaths: A list of XPaths of the elements to be clicked.
    :param name: The name to give this lambda function.
    :return: A lambda function that takes a driver and performs the clicking actions.
    """
    def _click_(driver):
        for xpath in xpaths:
            driver.click(xpath)
    _click_.__name__ = name
    return _click_


def _shortest_path(start: State, destination: State | str) -> list[Transition] | None:
    """
       Finds the shortest path between two states.

       This function uses a breadth-first search algorithm to find the shortest path
       from the starting state to the destination state. The destination can be specified
       either as a State object or as a string representing the name of the state.

       :param start: The starting state for the path search.
       :param destination: The destination state or state name for the path search.
       :return: A list of transitions representing the shortest path from the start
                state to the destination state. Returns None if no path is found.
       """
    visited = set()
    queue = deque([(start, [])])
    while queue:
        state, path = queue.popleft()
        # if this is a path to the desired state, return the path
        if state == destination or state.id == destination:
            return path
        # we do not want cycles: skip paths to already visited states
        if state in visited:
            continue
        visited.add(state)
        # take a step in all possible directions
        for transition in state.transitions:
            queue.append((transition.to_state, path + [transition]))
    return None

class TransitionError(Exception):
    """
    Exception raised when there is an error in state transition.
    """
    pass



import re
from time import sleep
from typing import Union, List

from puma.apps.android.whatsapp import logger
from puma.apps.android.whatsapp.states import *
from puma.apps.android.whatsapp.xpaths import *
from puma.state_graph.action import action
from puma.state_graph.puma_driver import PumaDriver, PumaClickException, supported_version
from puma.state_graph.state import SimpleState, compose_clicks
from puma.state_graph.state_graph import StateGraph
from puma.utils.xpath_utils import build_content_desc_xpath_widget


def go_to_chat(driver: PumaDriver, conversation: str):
    """
    Navigates to a specific chat conversation in the application.

    This function constructs an XPath to locate and click on a conversation element
    based on the conversation name. It is designed to be used within a state transition
    to navigate to a specific chat state.

    :param driver: The PumaDriver instance used to interact with the application.
    :param conversation: The name of the conversation to navigate to.
    """
    driver.get_elements(CONVERSATIONS_ROW_BY_SUBJECT.format(conversation=conversation))[-1].click()


def go_to_voice_call(driver: PumaDriver, contact: str):
    """
    Starts a voice call with a specific user.

    :param driver: The PumaDriver instance used to interact with the application.
    :param contact: The name of user to call.
    """
    driver.click(CALL_TAB_SEARCH_BUTTON)
    driver.send_keys(SEARCH_BAR, contact)
    driver.click(VOICE_CALL_START_BUTTON)


def go_to_video_call(driver: PumaDriver, contact: str):
    """
    Starts a video call with a specific user.

    :param driver: The PumaDriver instance used to interact with the application.
    :param contact: The name of user to call.
    """
    driver.click(CALL_TAB_SEARCH_BUTTON)
    driver.send_keys(SEARCH_BAR, contact)
    driver.click(VIDEO_CALL_START_BUTTON)


@supported_version("2.25.31.76")
class WhatsApp(StateGraph):
    """
    A class representing a state graph for managing UI states and transitions in the WhatsApp application.

    This class uses a state machine approach to manage transitions between different states
    of the WhatsApp user interface. It provides methods to navigate between states, validate states,
    and handle unexpected states or errors.
    """

    conversations_state = SimpleState([CONVERSATIONS_WHATSAPP_LOGO,
                                       CONVERSATIONS_HOME_ROOT_FRAME,
                                       CONVERSATIONS_NEW_CHAT_OR_SEND_MESSAGE,
                                       CONVERSATIONS_CHAT_TAB],
                                      initial_state=True)
    settings_state = SimpleState([SETTINGS_QR,
                                  SETTINGS_ACCOUNT_SWITCH],
                                 parent_state=conversations_state)
    profile_state = SimpleState([PROFILE_PROFILE_PICTURE,
                                 PROFILE_NAME,
                                 PROFILE_PHONE],
                                parent_state=settings_state)
    chat_state = WhatsAppChatState(parent_state=conversations_state)
    new_chat_state = SimpleState([NEW_CHAT_HEADER,
                                  NEW_CHAT_NEW_GROUP,
                                  NEW_CHAT_NEW_CONTACT,
                                  NEW_CHAT_NEW_COMMUNITY],
                                 parent_state=conversations_state)
    calls_state = SimpleState([CALLS_HEADER,
                               CALLS_START_CALL],
                              parent_state=conversations_state)
    updates_state = SimpleState([UPDATES_HEADER,
                                 UPDATES_STATUS_HEADER,
                                 UPDATES_NEW_STATUS],
                                parent_state=conversations_state)
    voice_call_state = WhatsAppVoiceCallState(parent_state=calls_state)
    video_call_state = WhatsAppVideoCallState(parent_state=calls_state)
    send_location_state = SimpleState([SEND_LOCATION_HEADER,
                                       SEND_LOCATION_LIVE_LOCATION,
                                       SEND_LOCATION_CURRENT_LOCATION],
                                      parent_state=chat_state)
    chat_settings_state = WhatsAppChatSettingsState(parent_state=chat_state)

    conversations_state.to(chat_state, go_to_chat)
    conversations_state.to(settings_state, compose_clicks([HAMBURGER_MENU, OPEN_SETTINGS_BY_TITLE],
                                                          name='navigate_to_settings'))
    conversations_state.to(new_chat_state, compose_clicks([CONVERSATIONS_NEW_CHAT_OR_SEND_MESSAGE],
                                                          name='press_new_chat_button'))
    conversations_state.to(calls_state, compose_clicks([CALLS_TAB], name='press_calls_tab'))
    conversations_state.to(updates_state, compose_clicks([UPDATES_TAB], name='press_updates_tab'))
    calls_state.to(voice_call_state, go_to_voice_call)
    calls_state.to(video_call_state, go_to_video_call)
    settings_state.to(profile_state, compose_clicks([PROFILE_INFO], name='press_profile'))
    chat_state.to(send_location_state, compose_clicks([CHAT_ATTACH_BUTTON, CHAT_ATTACH_LOCATION_BUTTON],
                                                      name='navigate_to_location'))
    chat_state.to(chat_settings_state, WhatsAppChatState.open_chat_settings)


    def __init__(self, device_udid: str, app_package: str):
        StateGraph.__init__(self, device_udid, app_package)

    def _handle_mention(self, message:str):
        """
        Make sure to convert a @name to an actual mention. Only one mention is allowed.
        :param message: The message containing the mention.
        """
        self.driver.send_keys(TEXT_ENTRY, message)
        sleep(1)
        # Find the mentioned name in the message. Note that it will search until the last word character. This means for
        # @jan-willem or @jan willem, only @jan will be found.
        mention_match = re.search(r"@\w+", message)
        mentioned_name = mention_match.group(0).strip("@")

        self.driver.press_left_arrow()
        while not (mention_suggestions := self.driver.get_elements(CHAT_MENTION_SUGGESTIONS)):
             self.driver.press_left_arrow()

        mentioned_person_el = \
            [person for person in
             mention_suggestions
             if mentioned_name.lower() in person.tag_name.lower()][0]
        mentioned_person_el.click()

        # Remove a space resulting from selecting the mention person
        self.driver.press_backspace()

    def _ensure_message_sent(self, message_text: str):
        message_status_el = self.driver.get_element(
            f"//*[@resource-id='com.whatsapp:id/conversation_text_row']"
            f"//*[@text='{message_text}']"  # Text field element containing message text
            f"/.."  # Parent of the message (i.e. conversation text row)
            f"//*[@resource-id='com.whatsapp:id/status']")  # Status element
        while message_status_el.tag_name == "Pending":
            self.gtl_logger.info("Message pending, waiting for the message to be sent.")
            sleep(10)
        return message_status_el

    @action(chat_state)
    def send_message(self, message_text: str, conversation: str = None):
        """
        Send a message in the current chat. If the message contains a mention, this is handled correctly.
        :param message_text: The text that the message contains.
        :param conversation: The chat conversation in which to send this message. Optional: not needed when already in a conversation
        """
        self.driver.click(TEXT_ENTRY)
        self._handle_mention(message_text) \
            if "@" in message_text \
            else self.driver.send_keys(TEXT_ENTRY, message_text)

        # Allow time for the link preview to load
        if 'http' in message_text:
            sleep(2)
        self.driver.click(SEND_CONTENT)

    @action(profile_state)
    def change_profile_picture(self, photo_dir_name: str, index: int = 1):
        self.driver.click(PROFILE_INFO_EDIT_BUTTON)
        self.driver.click(PROFILE_GALLERY)
        self.driver.click(PROFILE_FOLDERS)
        self._find_media_in_folder(photo_dir_name, index)
        self.driver.click(OK_BUTTON)

    @action(updates_state)
    def add_status(self, caption: str = None):
        """
        Sets a status by taking a picture and setting the given caption.
        Note: The first time an update is created, a pop-up will appear to change your privacy settings. This has to be
        handled once manually.
        :param caption: the caption to publish with the status.
        """
        self.driver.click(UPDATES_NEW_STATUS)
        self.driver.click(UPDATES_CAMERA_BUTTON)
        self.driver.click(UPDATES_SHUTTER)
        if caption:
             self.driver.send_keys(UPDATES_EDIT_CAPTION, caption)
        self.driver.click(SEND_RESOURCE)

    @action(profile_state)
    def set_about(self, about_text: str):
        self.driver.click(PROFILE_INFO_STATUS_CARD)
        self.driver.click(PROFILE_STATUS_EDIT_ICON)
        self.driver.send_keys(EDIT_TEXT, about_text)
        self.driver.click(PROFILE_SAVE_BUTTON)
        # This action ends in a screen that isn't a state, so move back one screen.
        self.driver.back()

    @action(new_chat_state)
    def create_new_chat(self, conversation: str, first_message: str):
        """
        Start a new 1-on-1 conversation with a contact and send a message.
        :param conversation: Contact to start the conversation with.
        :param first_message: First message to send to the contact
        """
        self.driver.click(f'//*[@resource-id="{WHATSAPP_PACKAGE}:id/contactpicker_text_container"]//*[@text="{conversation}"]')
        self.driver.click(build_content_desc_xpath_widget('Button', 'Message'))
        self.send_message(first_message)

    def open_more_options(self):
        """
        Open more options (hamburger menu) in the home screen.
        """
        self.driver.click(HAMBURGER_MENU)

    @action(conversations_state)
    def send_broadcast(self, receivers: List[str], broadcast_text: str):
        """
        Broadcast a message.
        :param receivers: list of receiver names, minimum of 2!
        :param broadcast_text: Text to send.
        """
        if len(receivers) < 2:
            raise Exception(f"Error: minimum of 2 receivers required for a broadcast, got: {receivers}")

        self.open_more_options()
        self.driver.click(CONVERSATIONS_NEW_BROADCAST_TITLE)
        for receiver in receivers:
            self.driver.click(CONVERSATIONS_CHAT_ABLE_CONTACT)

        self.driver.click(NEXT_BUTTON)
        self.driver.send_keys(TEXT_ENTRY, broadcast_text)
        self.driver.click(SEND_RESOURCE)

    @action(chat_state)
    def delete_message_for_everyone(self, conversation: str, message_text: str):
        """
        Remove a message with the message text. Should be recently sent, so it is still in view and still possible to
        delete for everyone.
        :param conversation: The chat conversation in which to send this message, if not currently in the desired chat.
        :param message_text: literal message text of the message to remove. The first match will be removed in case
        there are multiple with the same text.
        """
        self.driver.long_click_element(f"//*[@resource-id='{WHATSAPP_PACKAGE}:id/conversation_text_row']//*[@text='{message_text}']")
        self.driver.click(CHAT_DELETE_BUTTON)
        self.driver.click(CHAT_DELETE_FOR_EVERYONE)

    @action(new_chat_state)
    def create_group(self, conversation: str, members: Union[str, List[str]]):
        """
        Create a new group.
        :param conversation: The subject of the group.
        :param members: The contact(s) you want to add to the group (string or list).
        """
        self.driver.click(NEW_CHAT_NEW_GROUP)

        members = [members] if not isinstance(members, list) else members
        for member in members:
            contacts = self.driver.get_elements(TEXT_VIEWS)
            member_to_add = [contact for contact in contacts if contact.text.lower() == member.lower()][0]
            member_to_add.click()

        self.driver.click(NEXT_BUTTON)
        self.driver.send_keys(CONVERSATIONS_GROUP_NAME, conversation)
        self.driver.click(OK_BUTTON)
        # Creating a group takes a few seconds
        sleep(2)

    @action(conversations_state)
    def archive_conversation(self, conversation: str):
        """
        Archives a given conversation.
        :param conversation: The conversation to archive.
        """
        self.driver.long_click_element(f'//*[contains(@resource-id,"{WHATSAPP_PACKAGE}:id/conversations_row_contact_name") and @text="{conversation}"]')
        self.driver.click(CONVERSATIONS_MENUITEM_ARCHIVE)
        # Wait until the archive popup disappeared
        archived_popup_present = True
        tries = 0
        while archived_popup_present and tries < 5:
            logger.info("Waiting for archived popup to disappear")
            sleep(5)
            tries += 1
            archived_popup_present = 'archived' in self.driver.get_elements(CONVERSATIONS_ARCHIVED)[0].text
        logger.info("Archive pop-up gone!")

    @action(chat_state)
    def open_view_once_photo(self, conversation: str):
        """
        Open view once photo in the specified chat. Should be done right after the photo is sent, to ensure the correct
        photo is opened, this will be the lowest one.
        :param conversation: The chat in which the photo has to be opened
        """
        self.driver.get_elements(CHAT_VIEW_ONCE_MEDIA)[-1].click()

    @action(chat_settings_state)
    def set_group_description(self, conversation: str, description: str):
        """
        Set the group description.
        :param conversation: Name of the group to set the description for.
        :param description: Description of the group.
        """
        self.driver.swipe_to_click_element(f'{build_wa_resource_id_xpath("no_description_view")} | {build_wa_resource_id_xpath("has_description_view")}')
        self.driver.send_keys(EDIT_TEXT, description)
        self.driver.click(OK_BUTTON)

    @action(chat_settings_state)
    def delete_group(self, conversation: str):
        """
        Leaves and deletes a given group.
        Assumes the group exists and hasn't been left yet.
        :param conversation: the group to be deleted.
        """
        self.leave_group(conversation)
        self.driver.click(CHAT_SETTINGS_CONTAINS_DELETE_GROUP)
        self.driver.click(CHAT_SETTINGS_CONTAINS_DELETE_GROUP)

    @action(chat_state)
    def reply_to_message(self, message_to_reply_to: str, reply_text: str, conversation: str = None):
        """
        Reply to a message.
        :param conversation: The chat conversation in which to send this message.
        :param message_to_reply_to: message you want to reply to.
        :param reply_text: message text you are sending in your reply.
        """
        message_xpath = f'//android.widget.TextView[@resource-id="{WHATSAPP_PACKAGE}:id/message_text" and contains(@text, "{message_to_reply_to}")]'
        self.driver.swipe_to_find_element(message_xpath)
        self.driver.long_click_element(message_xpath)
        self.driver.click(CHAT_REPLY)
        self.driver.send_keys(TEXT_ENTRY, reply_text)
        self.driver.click(SEND)

    @action(chat_state)
    def send_emoji(self, conversation: str):
        """
        Send the first emoji in the emoji menu.
        :param conversation: The chat conversation in which to send this sticker.
        """
        self.driver.click(CHAT_EMOJI_PICKER)
        sleep(1)
        self.driver.click(CHAT_EMOJIS)
        self.driver.click(CHAT_EMOJI)
        self.driver.click(SEND)

    @action(chat_state)
    def send_sticker(self, conversation: str = None):
        """
        Send the first sticker in the sticker menu.
        :param conversation: The chat conversation in which to send this sticker.
        """
        self.driver.click(CHAT_EMOJI_PICKER)
        sleep(1)
        self.driver.click(CHAT_STICKERS)
        self.driver.click(CHAT_STICKER)

    @action(chat_state)
    def send_voice_message(self, duration: int = 2, conversation: str = None):
        """
        Sends a voice message in the specified conversation.
        :param conversation: The chat conversation in which to send this voice recording.
        :param duration: the duration in of the voice message to send in seconds.
        """
        self.driver.long_click_element(CHAT_VOICE_NOTE_BUTTON, duration=duration)

    @action(send_location_state, end_state=chat_state)
    def send_current_location(self, conversation: str = None):
        """
        Send the current location in the specified chat.
        :param conversation: The chat conversation in which to send the location.
        """
        sleep(5)  # it takes some time to fix the location
        self.driver.click(SEND_LOCATION_CURRENT_LOCATION_BUTTON)

    @action(send_location_state, end_state=chat_state)
    def send_live_location(self, caption: str = None, conversation: str = None):
        """
        Send a live location in the specified chat.
        :param conversation: The chat conversation in which to start the live location sharing.
        :param caption: Optional caption sent along with the live location
        """
        self.driver.click(SEND_LOCATION_LIVE_LOCATION_BUTTON)
        if self.driver.is_present(SEND_LOCATION_LIVE_LOCATION_DIALOG):
            self.driver.click(SEND_LOCATION_POPUP_CONTINUE)
        if caption is not None:
            self.driver.send_keys(SEND_LOCATION_CAPTION, caption)
        self.driver.click(SEND)

    @action(chat_state)
    def stop_live_location(self, conversation: str = None):
        """
        Stops the current live location sharing.
        :param conversation: The chat conversation in which to stop the live location sharing.
        """
        self.driver.swipe_to_click_element(CHAT_STOP_SHARING)

        if self.driver.is_present(STOP_BUTTON):
            self.driver.click(STOP_BUTTON)

    @action(chat_state)
    def send_contact(self, contact_name: str, conversation: str = None):
        """
        Send a contact in the specified chat.
        :param contact_name: the name of the contact to send.
        :param conversation: The chat conversation in which to send the contact.
        """
        self.driver.click(CHAT_ATTACH_BUTTON)
        self.driver.click(CHAT_ATTACH_CONTACT_BUTTON)
        self.driver.swipe_to_click_element(CHAT_CONTACT_NAME.format(contact_name=contact_name))
        self.driver.click(NEXT_BUTTON)
        self.driver.click(CHAT_ATTACH_SEND_BUTTON)

    @action(chat_settings_state)
    def activate_disappearing_messages(self, conversation: str = None):
        """
        Activates disappearing messages (auto delete) in the current or a given chat.
        Messages will now auto-delete after 24h.
        :param conversation: The conversation for which disappearing messages should be activated.
        """
        self.driver.swipe_to_click_element(CHAT_SETTINGS_DISAPPEARING_MESSAGES)
        self.driver.click(RADIO_BUTTON_24_HOURS)
        self.driver.back()

    @action(chat_settings_state)
    def deactivate_disappearing_messages(self, conversation: str = None):
        """
        Disables disappearing messages (auto delete) in the current or a given chat.
        :param conversation: The conversation for which disappearing messages should be activated.
        """
        self.driver.swipe_to_click_element(CHAT_SETTINGS_DISAPPEARING_MESSAGES)
        self.driver.click(RADIO_BUTTON_OFF)
        self.driver.back()

    @action(calls_state, end_state=voice_call_state)
    def start_voice_call(self, conversation: str):
        """
        Make a WhatsApp voice call. The call is made to a given contact.
        :param conversation: name of the contact to call.
        """
        go_to_voice_call(self.driver, conversation)

    @action(calls_state, end_state=video_call_state)
    def start_video_call(self, conversation: str):
        """
        Make a WhatsApp voice call. The call is made to a given contact.
        :param conversation: name of the contact to call.
        """
        go_to_video_call(self.driver, conversation)

    @action(voice_call_state, end_state=calls_state)
    def end_voice_call(self, conversation: str = None):
        """
        Ends the current voice call.
        """
        self._end_call()

    @action(video_call_state, end_state=calls_state)
    def end_video_call(self, conversation: str = None):
        """
        Ends the current video call.
        """
        self._end_call()

    def _end_call(self):
        if not self.driver.is_present(CALL_END_CALL_BUTTON, implicit_wait=1):
            # tap screen to make call button visible
            self.driver.click(CALL_SCREEN_BACKGROUND)
        self.driver.click(CALL_END_CALL_BUTTON)
        # Go back twice to ensure we are back in a recognized state.
        self.driver.back()
        self.driver.back()

    # This method is not an @action, since it is not tied to a state.
    def answer_call(self):
        """
        Answer when receiving a call via Whatsapp.
        """
        self.driver.open_notifications()
        sleep(2)
        self.driver.click(RECEIVE_CALL_ANSWER_BUTTON)

    # This method is not an @action, since it is not tied to a state.
    def decline_call(self):
        """
        Declines an incoming Whatsapp call.
        """
        self.driver.open_notifications()
        sleep(2)
        self.driver.click(RECEIVE_CALL_DECLINE_BUTTON)

    @action(chat_settings_state)
    def leave_group(self, conversation: str = None):
        """
        This method will leave the given group. It will not delete that group.
        :param conversation: Name of the group we want to leave.
        """
        self.driver.swipe_to_click_element(CHAT_SETTINGS_EXIT_GROUP_LIST_ITEM)
        self.driver.click(CHAT_SETTINGS_EXIT_GROUP_BUTTON)

    @action(chat_settings_state)
    def remove_member_from_group(self, conversation: str, member: str):
        """
        Removes a given member from a given group chat.
        It is assumed the group chat exists and has the given member.
        :param conversation: The group
        :param member: The member to remove
        """
        self.driver.swipe_to_click_element(CHAT_SETTINGS_MEMBER.format(member=member))
        self.driver.click(CHAT_SETTINGS_REMOVE_MEMBER)
        self.driver.click(CHAT_SETTINGS_OK_BUTTON)

    @action(chat_state)
    def forward_message(self, conversation: str, message_contains: str, to_chat: str):
        """
        Forwards a message from one conversation to another.
        It is assumed the message and both conversations exists.
        :param conversation: The chat from which the message has to be forwarded
        :param message_contains: the text from the message that has to be forwarded. Uses String.contains(), so only part
        of the message is needed, but be sure the given text is enough to match your intended message uniquely.
        :param to_chat: The chat to which the message has to be forwarded.
        """
        self.driver.long_click_element(CHAT_MESSAGE_BY_CONTENT.format(message_contains=message_contains))
        self.driver.click(CHAT_FORWARD_MESSAGE)
        self.driver.click(CHAT_FORWARD_CONTACT_BY_NAME.format(to_chat=to_chat))
        self.driver.click(SEND)

    @action(chat_state)
    def send_media(self, directory_name: str, conversation: str = None, index: int = 1, caption: str = None,
                   view_once: bool = False):
        # Go to gallery
        self.driver.click(CHAT_ATTACH_BUTTON)
        self.driver.click(CHAT_ATTACH_GALLERY_BUTTON)
        self.driver.click(CHAT_GALLERY_FOLDERS_BUTTON)
        self._find_media_in_folder(directory_name, index)
        sleep(0.5)
        self.driver.click(CHAT_FIRST_MEDIA_IN_FOLDER)

        if caption:
            sleep(0.5)
            self.driver.send_keys(CHAT_CAPTION_TEXT_BOX, caption)
            # Clicking the text box after sending keys is required for Whatsapp to notice text has been inserted.
            self.driver.click(CHAT_CAPTION_TEXT_BOX)
            self.driver.back()

        if view_once:
            self.driver.click(CHAT_SEND_MEDIA_VIEW_ONCE)
            if self.driver.is_present(CHAT_POPUP_BUTTON_OK):
                self.driver.click(CHAT_POPUP_BUTTON_OK)
        sleep(1)
        self.driver.click(SEND)

    def _find_media_in_folder(self, directory_name: str, index: int):
        try:
            self.driver.swipe_to_click_element(CHAT_DIRECTORY_NAME.format(directory_name=directory_name))
        except PumaClickException:
            raise PumaClickException(f'The directory {directory_name} could not be found.')
        self.driver.click(CHAT_DIRECTORY_NAME.format(directory_name=directory_name))
        sleep(0.5)
        try:
            self.driver.click(CHAT_DIRECTORY_MEDIA_BY_INDEX.format(index=index))
        except PumaClickException:
            raise PumaClickException(
                f'The media at index {index} could not be found. The index is likely too large or negative.')

    def is_message_marked_sent(self, message_text: str, implicit_wait: float = 5):
        """
        Verify that a message with given text has been sent in the current conversation.

        The message must be visible in the current chat view. It must also be marked specifically as sent,
        i.e. a single uncoloured (grey) checkmark.

        :param message_text: the text of the message which should have been sent
        :param implicit_wait: the maximum time to wait for a message to be marked sent, in seconds
        :return: True if the expected message is marked as sent, False otherwise
        """
        sent_message_xpath = CHAT_MESSAGE_BY_CONTENT_AND_STATE.format(message_text=message_text, state='Sent')

        return self.driver.is_present(sent_message_xpath, implicit_wait=implicit_wait)

    def is_message_marked_delivered(self, message_text: str, implicit_wait: float = 5):
        """
        Verify that a message with given text has been delivered in the current conversation.

        The message must be visible in the current chat view. It must also be marked specifically as delivered,
        i.e. two uncoloured (grey) checkmarks.

        :param message_text: the text of the message which should have been delivered
        :param implicit_wait: the maximum time to wait for a message to be marked delivered, in seconds
        :return: True if the expected message is marked as delivered, False otherwise
        """
        delivered_message_xpath = CHAT_MESSAGE_BY_CONTENT_AND_STATE.format(message_text=message_text, state='Delivered')

        return self.driver.is_present(delivered_message_xpath, implicit_wait=implicit_wait)

    def is_message_marked_read(self, message_text: str, implicit_wait: float = 10):
        """
        Verify that a message with given text has been read in the current conversation.

        The message must be visible in the current chat view. It must also be marked specifically as read,
        i.e. two (blue) coloured checkmarks.

        :param message_text: the text of the message which should have been read
        :param implicit_wait: the maximum time to wait for a message to be marked read, in seconds
        :return: True if the expected message is marked as read, False otherwise
        """
        read_message_xpath = CHAT_MESSAGE_BY_CONTENT_AND_STATE.format(message_text=message_text, state='Read')

        return self.driver.is_present(read_message_xpath, implicit_wait=implicit_wait)

    def in_connected_call(self, implicit_wait: float = 5):
        """
        Verify that we are in a connected call. This can be either a voice call or a video call.

        :param implicit_wait: the maximum time to wait until a call is connected, in seconds
        :return: True if we are in an active call, False otherwise
        """
        # first check if we see the end call button
        if not self.driver.is_present(CALL_END_CALL_BUTTON, implicit_wait=implicit_wait):
            # if not, tap screen to (try and) make call button visible
            self.driver.click(CALL_SCREEN_BACKGROUND)

        # now check again if we see a call end button; not using the passed
        # implicit wait, since we already waited, and it should not take more than a second to make
        # the call button visible after touch
        return self.driver.is_present(CALL_END_CALL_BUTTON, implicit_wait=1)

    def group_exists(self, conversation: str, members: Union[str, List[str]]):
        """
        Verify that a group exists with given name and members. Will log a warning when
        the expected group can't be found, or doesn't contain expected members.

        You should currently be in the 'conversations_state' state, i.e. the WhatsApp initial state.
        Will wait a short amount of time if necessary for the group to be created.

        :param conversation: the expected name of the group
        :param members: the expected members of the group
        :return: True if the expected group with given members exists, False otherwise
        """
        expected_member_names = members if isinstance(members, list) else [members]

        try:
            go_to_chat(self.driver, conversation)
        except PumaClickException as e:
            self.gtl_logger.warning(f'Failed to find group with name {conversation}: {str(e)}')
            return False
        else:
            WhatsAppChatState.open_chat_settings(self.driver, conversation)

            found_member_elements = self.driver.swipe_to_find_elements(CHAT_SETTINGS_ANY_MEMBER, max_swipes=4)
            found_member_names = [member.text for member in found_member_elements]

            if set(found_member_names) != set(expected_member_names):
                self.gtl_logger.warning(f"Group with name '{conversation}' does not contain expected members:"
                               f" expected {sorted(expected_member_names)},"
                               f" actual {sorted(found_member_names)}")
                return False

        return True



from time import sleep

from puma.state_graph.action import action
from puma.state_graph.popup_handler import PopUpHandler
from puma.state_graph.puma_driver import supported_version
from puma.state_graph.state_graph import StateGraph
from puma.state_graph.state import SimpleState, compose_clicks


PHOTO_STATE_TAKE_PHOTO = '//android.widget.ImageButton[@content-desc="Take photo"]'
PHOTO_STATE_CAMERA = '//android.widget.TextView[@content-desc="Camera"]'
PHOTO_STATE_SWITCH_TO_VIDEO = '//android.widget.TextView[@content-desc="Switch to Video Camera"]'
PHOTO_STATE_SHUTTER_BUTTON = '//android.widget.ImageButton[@resource-id="com.google.android.GoogleCamera:id/shutter_button"]'

VIDEO_STATE_VIDEO = '//android.widget.TextView[@content-desc="Video"]'
VIDEO_STATE_TAKE_VIDEO = '//android.widget.ImageButton[@content-desc="Start video"]'
VIDEO_STATE_SWITCH_TO_PHOTO = '//android.widget.TextView[@content-desc="Switch to Camera Mode"]'
VIDEO_STATE_STOP_VIDEO = '//android.widget.ImageButton[@content-desc="Stop video"]'

SWITCH_CAMERA_BUTTON = '//android.widget.ImageButton[@resource-id="com.google.android.GoogleCamera:id/camera_switch_button"]'

SETTINGS_STATE_CAMERA_SETTINGS = '//android.widget.TextView[@text="Camera settings"]'
SETTINGS_STATE_TITLE = '//android.widget.TextView[@resource-id="android:id/title" and @text="General"]'

OPEN_OPTIONS_MENU = '//android.widget.ImageView[@content-desc="Open options menu"]'
OPEN_SETTINGS = '//android.widget.Button[@content-desc="Open settings"]'

POPUP_TURN_ON_BY_DEFAULT = '//android.widget.TextView[@text="Turned on by default"]'
POPUP_BOTTOMSHEET = '//android.widget.LinearLayout[@resource-id="com.google.android.GoogleCamera:id/bottomsheet_container"]'
POPUP_GOT_IT = '//android.widget.Button[@resource-id="com.google.android.GoogleCamera:id/got_it_button"]'
DONE = '//android.widget.Button[@text="Done"]'

APPLICATION_PACKAGE = 'com.google.android.GoogleCamera'


@supported_version("8.8.225.510547499.09")
class GoogleCamera(StateGraph):
    """
    A class representing a state graph for managing UI states and transitions in the Google Camera application.

    This class uses a state machine approach to manage transitions between different states
    of the Google Camera user interface. It provides methods to navigate between states,
    take pictures, and record videos.
    """

    # Define states
    photo = SimpleState(xpaths=[PHOTO_STATE_TAKE_PHOTO,
                                PHOTO_STATE_CAMERA],
                        initial_state=True)
    video = SimpleState(xpaths=[VIDEO_STATE_VIDEO,
                                VIDEO_STATE_TAKE_VIDEO])
    settings = SimpleState(xpaths=[SETTINGS_STATE_CAMERA_SETTINGS,
                                   SETTINGS_STATE_TITLE],
                           parent_state=photo)  # note that settings does not have a real parent state. the back restores the last state before navigating to settings.

    # Define transitions. Only forward transitions are needed, back transitions are added automatically
    photo.to(video, compose_clicks([PHOTO_STATE_SWITCH_TO_VIDEO], name='go_to_video'))
    video.to(photo, compose_clicks([VIDEO_STATE_SWITCH_TO_PHOTO], name='go_to_camera'))
    go_to_settings = compose_clicks([OPEN_OPTIONS_MENU, OPEN_SETTINGS], name= 'go_to_settings')
    photo.to(settings, go_to_settings)
    video.to(settings, go_to_settings)

    def __init__(self, device_udid):
        """
        Initializes the GoogleCamera with a device UDID.

        :param device_udid: The unique device identifier for the Android device.
        """
        StateGraph.__init__(self, device_udid, APPLICATION_PACKAGE)
        self.add_popup_handler(PopUpHandler([POPUP_TURN_ON_BY_DEFAULT], [DONE]))
        self.add_popup_handler(PopUpHandler([POPUP_BOTTOMSHEET], [POPUP_GOT_IT]))

    @action(photo)
    def take_picture(self, front_camera: bool = None):
        """
        Takes a single picture.

        This method ensures the correct camera view (front or back) and then takes a picture.

        :param front_camera: If True, uses the front camera; if False, uses the back camera; if None, no change is made.
        """
        self._ensure_correct_camera_view(front_camera)
        self.driver.click(PHOTO_STATE_SHUTTER_BUTTON)

    @action(video)
    def record_video(self, duration, front_camera: bool = None):
        """
        Records a video for the given duration.

        This method ensures the correct camera view (front or back) and then starts and stops video recording.

        :param duration: The duration in seconds to record the video.
        :param front_camera: If True, uses the front camera; if False, uses the back camera; if None, no change is made.
        """
        self._ensure_correct_camera_view(front_camera)
        self.driver.click(VIDEO_STATE_TAKE_VIDEO)
        sleep(duration)
        self.driver.click(VIDEO_STATE_STOP_VIDEO)

    def _ensure_correct_camera_view(self, front_camera):
        """
        Ensures the correct camera view (front or back) is selected.

        :param front_camera: If True, ensures the front camera is selected; if False, ensures the back camera is selected.
        """
        if front_camera is None:
            return
        switch_button = self.driver.get_element(SWITCH_CAMERA_BUTTON)
        currently_in_front = 'front' not in switch_button.get_attribute("content-desc")
        if currently_in_front != front_camera:
            switch_button.click()



CHAT_OVERVIEW_NEW_MESSAGE_BUTTON = '//android.widget.FrameLayout[@content-desc="New Message"]'
CHAT_OVERVIEW_NAV_MENU_BUTTON = '//android.widget.ImageView[@content-desc="Open navigation menu"]'
CHAT_OVERVIEW_SEARCH_BUTTON = '//android.widget.ImageButton[@content-desc="Search"]'

SEARCH_INPUT_FIELD = '//android.widget.EditText[@text="Search"]'
FIRST_SEARCH_HIT = '//androidx.recyclerview.widget.RecyclerView/android.view.ViewGroup[starts-with(lower-case(@text), lower-case("{conversation},"))]'

CHAT_STATE_CONVERSATION_NAME = '//android.widget.ImageView[@content-desc="Go back"]/../android.widget.FrameLayout[starts-with(lower-case(@content-desc),lower-case("{conversation}\n"))]/android.widget.TextView[starts-with(lower-case(@text), lower-case("{conversation}"))]'
CHAT_STATE_CALL_BUTTON = '//android.widget.ImageButton[@content-desc="Call"]'
CHAT_STATE_BACK_BUTTON = '//android.widget.ImageView[@content-desc="Go back"]'
CHAT_STATE_THREE_DOTS = '//android.widget.ImageButton[@content-desc="More options"]'
CHAT_STATE_GIF_BUTTON = '//android.widget.ImageView[@content-desc="Emoji, stickers, and GIFs"]'
CHAT_STATE_MESSAGE_TEXTBOX = '//android.widget.EditText[@text]'
CHAT_STATE_MEDIA_BUTTON = '//android.widget.ImageView[@content-desc="Attach media"]'
CHAT_STATE_RECORD_VIDEO_OR_AUDIO_MESSAGE = '//android.widget.FrameLayout[@content-desc="Record voice message" or @content-desc="Record video message"]'
CHAT_STATE_SEND_BUTTON = '//android.view.View[@content-desc="Send"]'
CHAT_STATE_STOP_LIVE_LOCATION_SHARING_BUTTON = '//android.widget.ImageView[@content-desc="Stop sharing live location"]'
CHAT_STATE_STOP_LIVE_LOCATION_CONFIRM_BUTTON = '//android.widget.TextView[@text="Stop"]'
CHAT_STATE_DELETE_AND_LEAVE_GROUP = '//android.widget.TextView[@text="Delete and Leave Group"]'
CHAT_STATE_DELETE_AND_LEAVE_GROUP_FOR_ALL = '//android.widget.CheckBox[@text="Delete the group for all members"]'
CHAT_STATE_DELETE_AND_LEAVE_GROUP_CONFIRM_BUTTON = '//android.widget.TextView[@text="Delete Group" or @text="Delete chat"]'

SEND_MEDIA_STATE_INSTANT_CAMERA_BUTTON = '//android.widget.FrameLayout[@content-desc="Instant camera"]/preceding-sibling::android.widget.FrameLayout[last()]'
SEND_MEDIA_STATE_GALLERY_BUTTON = '//android.widget.FrameLayout[@text="Gallery"]'
SEND_MEDIA_STATE_FILE_BUTTON = '//android.widget.FrameLayout[@text="File"]'
SEND_MEDIA_STATE_LOCATION_BUTTON = '//android.widget.FrameLayout[@text="Location"]'

SEND_MEDIA_STATE_SEND_CURRENT_LOCATION_BUTTON = '//android.widget.TextView[@text="Send selected location"]'
SEND_MEDIA_STATE_MY_LOCATION_BUTTON = '//android.widget.ImageView[@content-desc="My location"]'
SEND_MEDIA_STATE_LIVE_LOCATION_BUTTON = '//android.widget.TextView[@text="Share My Live Location for..."]'
SEND_MEDIA_STATE_LIVE_LOCATION_DURATION_OPTION_INDEX = '(//android.widget.RadioButton/android.widget.TextView[@text])[{duration}]'
SEND_MEDIA_STATE_LIVE_LOCATION_DURATION_OPTION_TEXT = '//android.widget.RadioButton/android.widget.TextView[contains(lower-case(@text), lower-case("{duration}"))]'
SEND_MEDIA_STATE_LIVE_LOCATION_SHARE_BUTTON = '//android.widget.TextView[@text="Share"]'

SEND_FROM_GALLERY_STATE_BACK_BUTTON = '//android.widget.ImageView[@content-desc="Go back"]'
SEND_FROM_GALLERY_THREE_DOTS_BUTTON = '//android.widget.ImageButton[@content-desc="More options"]'
SEND_FROM_GALLERY_FOLDER_PICKER = '//android.widget.ImageView[@content-desc="Go back"]/following-sibling::*[1][@text]'
SEND_FROM_GALLERY_MEDIA_SWITCH = '(//android.widget.Switch)[{index}]'

CALL_STATE_END_CALL_BUTTON = '//android.widget.Button[@text="End Call"]'
CALL_STATE_SPEAKER_BUTTON = '//android.widget.FrameLayout[@content-desc="Speaker"]'
CALL_STATE_MUTE_BUTTON = '//android.widget.FrameLayout[@content-desc="Mute" or @content-desc="Unmute"]'
CALL_STATE_STATUS = '//android.widget.LinearLayout[ends-with(@text, "Telegram Call")]/android.widget.FrameLayout/android.widget.TextView'

NEW_MESSAGE_STATE_NEW_GROUP_BUTTON = '//android.widget.FrameLayout[@text="New Group"]'
NEW_MESSAGE_STATE_NEW_CONTACT_BUTTON = '//android.widget.FrameLayout[@text="New Contact"]'
NEW_MESSAGE_STATE_NEW_CHANNEL_BUTTON = '//android.widget.FrameLayout[@text="New Channel"]'
NEW_MESSAGE_STATE_CREATE_NEW_CONTACT_BUTTON = '//android.widget.FrameLayout[@content-desc="Create New Contact"]'

NEW_GROUP_MEMBER = '//android.widget.FrameLayout/android.widget.TextView[lower-case(@text)=lower-case("{member}")]'
NEW_GROUP_NEXT = '//android.widget.ImageView[@content-desc="Next"]'
NEW_GROUP_NAME_INPUT = '//android.widget.EditText'
NEW_GROUP_AUTO_DELETE_OPTION = '//android.widget.TextView[@text="Auto-Delete Messages"]'
NEW_GROUP_DONE_BUTTON = '//android.widget.FrameLayout[@content-desc="Done"]'

CHAT_SETTINGS_STATE_BACK = '//android.widget.ImageView[@content-desc="Go back"]'
CHAT_SETTINGS_STATE_THREE_DOTS = '//android.widget.ImageButton[@content-desc="More options"]'
CHAT_SETTINGS_STATE_CONVERSATION_NAME = '(//android.widget.TextView[@text and @text!=""])[2]'
CHAT_SETTINGS_STATE_CONVERSATION_NAME_CONTEXT = '(//android.widget.TextView[lower-case(@text)=lower-case("{conversation}")])[2]'
CHAT_SETTINGS_STATE_CONVERSATION_STATUS_CONTEXT = '(//android.widget.TextView[lower-case(@text)=lower-case("{conversation}")])[2]/following-sibling::*[1][@text and @text!=""]'
# group settings
CHAT_SETTINGS_EDIT_GROUP_BUTTON = '//android.widget.ImageButton[@content-desc="Edit"]'
CHAT_SETTINGS_STATE_ADD_MEMBERS = '//android.widget.TextView[@text="Add Members"]'
CHAT_SETTINGS_STATE_ADD_MEMBERS_SEARCH = '//android.widget.EditText[@text]'
CHAT_SETTINGS_STATE_MEMBER_CONTEXT = '//androidx.recyclerview.widget.RecyclerView/android.widget.FrameLayout[android.view.View and count(android.widget.TextView) >= 2 and android.widget.TextView[lower-case(@text)=lower-case("{member}")]]'
CHAT_SETTINGS_STATE_REMOVE_MEMBER_BUTTON = '//android.widget.TextView[@text="Remove from group"]'

EDIT_GROUP_NAME= '//android.widget.EditText[1]'
EDIT_GROUP_DESCRIPTION= '//android.widget.EditText[2]'
EDIT_GROUP_DONE= '//android.widget.ImageButton[@content-desc="Done"]'
